﻿using System;
using System.Threading;
using System.Threading.Tasks;
using System.Net.Http;
using System.Text;
using Websocket.Client; //nuget package
using Json.Net; //nuget package
using Cryptography.ECDSA; //nuget package Cryptography.ECDSA.Secp256k1
using Free.Ports.zLib; //To decompress incoming messages. Source code included under zlib.net folder. This code is from the excellent https://github.com/JoshDullen/zlib.net

namespace viking_trade
{
    class Message
    {
        public long T; //time epoch sec UTC
        public string P; //topic
        public string Y; //type
    }
    class Program
    {
        private const int OutBufSize = 32768;
        static async Task<long> getExchangeTime() {
            long exchangeTime = 0;
            HttpClient clientREST = new HttpClient();
            clientREST.BaseAddress = new Uri("https://api.viking.trade/v1/public/g_t");
            HttpResponseMessage response = clientREST.GetAsync("").Result;
            if (response.IsSuccessStatusCode)
            {
                var msgText = await response.Content.ReadAsStringAsync();
                var msg = JsonNet.Deserialize<Message>(msgText);
                exchangeTime = msg.T;
            }
            else
            {
                Console.WriteLine("{0} ({1})", (int)response.StatusCode, response.ReasonPhrase);
                Console.WriteLine("Error. Failed to get exchange time. Exiting");
                System.Environment.Exit(0);
            }
            clientREST.Dispose();
            return exchangeTime;
        }
        static async Task Main(string[] args) {
            //init zlib for incoming messages decompression
            var zs = new zlib.z_stream();
            if (zlib.Z_OK != zlib.inflateInit2(zs, -15)) {
                Console.WriteLine("Error. Zlib init failed. Exiting");
                System.Environment.Exit(0);
            }
            zs.out_buf = new byte[OutBufSize];
            zs.avail_out = OutBufSize;
            zs.next_out = 0;
            zs.total_out = 0;

            var exitEvent = new ManualResetEvent(false);
            var url = new Uri("wss://api.viking.trade/private");
            
            using (var client = new WebsocketClient(url))
            {
                client.ReconnectionHappened.Subscribe(async type => {
                    Console.WriteLine($"Connected to: {url}");
                    
                    var heartbeatTimer = new System.Timers.Timer();
                    heartbeatTimer.Interval = 5000;
                    heartbeatTimer.Elapsed += delegate (Object source, System.Timers.ElapsedEventArgs e) {
                        client.Send("7"); //heartbeat each 5 seconds
                    };
                    heartbeatTimer.Enabled = true;

                    long exchangeTime = await getExchangeTime();

                    Console.WriteLine($"Time: {exchangeTime}");

                    var key_id = "YOUR_API_KEY_ID_STRING";
                    var key_secret = "YOUR_API_KEY_SECRET_64_SYMBOLS_STRING";
                    var seckey = Hex.HexToBytes(key_secret);
                    var data = Sha256Manager.GetHash(Encoding.UTF8.GetBytes(key_id + " " + exchangeTime));
                    Console.WriteLine($"Sha256: {BitConverter.ToString(data).Replace("-", string.Empty)}");
                    var sigBytes = Secp256K1Manager.SignCompressedCompact(data, seckey);
                    var sig = BitConverter.ToString(sigBytes).Replace("-", string.Empty).Substring(2);
                    Console.WriteLine($"Sig: {sig}");
                    client.Send("{\"P\":\"auth_key\", \"D\":{\"kid\": \"" + key_id + "\", \"ts\": " + exchangeTime + ", \"sig\":\"" + sig + "\"}}");
                   
                });

                //Msg handler
                client.MessageReceived.Subscribe( msg => {
                    //Console.WriteLine($"Message received: {msg} {msg.MessageType}");
                    String text = null;
                    if (msg.MessageType == System.Net.WebSockets.WebSocketMessageType.Text) {
                        text = msg.Text;
                    } else
                    if (msg.MessageType == System.Net.WebSockets.WebSocketMessageType.Binary) {//decompress
                        var tail = new byte[4] { 0x00, 0x00, 0xFF, 0xFF };
                        var rv = new byte[msg.Binary.Length + tail.Length];
                        System.Buffer.BlockCopy(msg.Binary, 0, rv, 0, msg.Binary.Length);
                        System.Buffer.BlockCopy(tail, 0, rv, msg.Binary.Length, tail.Length);
                        zs.next_in = 0;
                        zs.avail_in = (uint) rv.Length;
                        zs.in_buf = rv;
                        zs.avail_out = OutBufSize;
                        zs.next_out = 0;
                        zlib.inflate(zs, zlib.Z_SYNC_FLUSH);
                        var res = new byte[zs.next_out];
                        System.Buffer.BlockCopy(zs.out_buf, 0, res, 0, zs.next_out);
                        text = System.Text.Encoding.Default.GetString(res);
                    } else
                    if (msg.MessageType == System.Net.WebSockets.WebSocketMessageType.Close) {
                        return;
                    }
                    if (text == "7") return; //skip heartbeat
                    Console.WriteLine($"Recieved: {text}");
                });

                await client.Start();

                exitEvent.WaitOne();
            }
        }
    }
}
